from scriptable.antlr.TypescriptParser import TypescriptParser
from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding


class Type(AST[str]):

    def __init__(self, value: str):
        self.value = value

    def execute(self, context: ASTBinding) -> str:
        return self.value

    @staticmethod
    def parse(ctx: TypescriptParser.STypeContext):
        return Type(ctx.getText())


from typing import List

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding
from scriptable.api.exit_value import GoTo


class While(AST[None]):

    def __init__(self, expression: AST, tail: AST):
        self.expression = expression
        self.tail = tail

    def execute(self, context: ASTBinding) -> None:
        counter = 0
        while self.expression.execute(context):
            result = self.tail.execute(context)

            if isinstance(result, GoTo) and result.value == "break":
                break
            if isinstance(result, GoTo) and result.value == "continue":
                continue

            assert counter <= context.restrictions.max_loops, "max loops exceeded"
            counter += 1
        return None

    @staticmethod
    def parse(branch: List[AST]) -> 'While':
        return While(branch[0], branch[1])



from typing import Any, List

from scriptable.api import AST
from scriptable.api.accessor import Accessor
from scriptable.api.ast_binding import ASTBinding


class PropertyAccess(AST[Any]):
    def __init__(self, branch: List[AST]):
        self.branch = branch

    def execute(self, binding: ASTBinding) -> bool:
        def unwrap(obj):
            if isinstance(obj, Accessor):
                return obj.value
            return obj

        branch = list(map(lambda ast: ast.execute(binding), self.branch))
        current = branch.pop(0)
        while len(branch) > 0:
            current = current[unwrap(branch.pop(0))]
        return current

    @staticmethod
    def parse(branch: List[AST]) -> 'PropertyAccess':
        return PropertyAccess(branch)



from typing import Any, List, Dict

from scriptable.api import AST
from scriptable.api.accessor import Accessor
from scriptable.api.ast_binding import ASTBinding
from scriptable.runtime.accessor.typescript.map_accessor import MapAccessor


class Map(AST[Accessor[Dict[str, Any]]]):
    def __init__(self, accessor: Accessor[Dict[str, Any]]):
        self.accessor = accessor

    def execute(self, binding: ASTBinding) -> Accessor[Dict[str, Any]]:
        return self.accessor

    @staticmethod
    def parse(branch: List[AST]) -> 'Map':
        branch = list(map(lambda x: x.execute(ASTBinding()), branch))
        _dict = {}
        for i in range(0, len(branch), 2):
            _dict[branch[i].value] = branch[i + 1]
        return Map(MapAccessor(_dict))



from typing import List

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding
from scriptable.api.exit_value import GoTo


class ForOf(AST[None]):

    def __init__(self, name: str, value: AST, tail: AST):
        self.name = name
        self.value = value
        self.tail = tail

    def execute(self, context: ASTBinding) -> None:
        if self.name in context.properties:
            raise ValueError(f"property {self.name} is already defined")

        value = self.value.execute(context)
        assert len(value) <= context.restrictions.max_loops, "max loops exceeded"

        if isinstance(value, str):
            for char in value:
                context.add_property(self.name, char)
                result = self.tail.execute(context)

                if isinstance(result, GoTo) and result.value == "break":
                    break
                if isinstance(result, GoTo) and result.value == "continue":
                    continue
        else:
            for item in value:
                context.add_property(self.name, item)
                result = self.tail.execute(context)

                if isinstance(result, GoTo) and result.value == "break":
                    break
                if isinstance(result, GoTo) and result.value == "continue":
                    continue

        return None

    @staticmethod
    def parse(name: str, branch: List[AST]) -> 'ForOf':
        # noinspection PyTypeChecker
        return ForOf(name, branch[0], branch[1])



from .Break import Break
from .Continue import Continue
from .For import For
from .ForIn import ForIn
from .ForOf import ForOf
from .While import While
from .ifelse import If
from .loop_tail import LoopTail



from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding


class MutableVar(AST[None]):
    def __init__(self, name: str, value: AST):
        self.name = name
        self.value = value

    def execute(self, binding: ASTBinding) -> None:
        binding.add_property(self.name, self.value.execute(binding))
        return None

    @staticmethod
    def parse(name: str, value: AST) -> 'MutableVar':
        return MutableVar(name, value)



from .function import Function
from .function_access import FunctionAccess
from .function_arg import FunctionArg
from .function_arg_def import FunctionArgDef
from .function_call import FunctionCall
from .function_head import FunctionHead
from .function_lambda import FunctionLambda
from .function_tail import FunctionTail


from typing import List

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding


class LogicExpression(AST[bool]):
    def __init__(self, branch: List[AST]):
        self.operand_stack = [branch[x] for x in range(0, len(branch), 2)]
        self.operator_stack = [branch[x] for x in range(1, len(branch), 2)]

    def execute(self, binding: ASTBinding) -> bool:
        operand_stack = list(map(lambda ast: ast.execute(binding), self.operand_stack))
        operator_stack = list(map(lambda ast: ast.execute(binding), self.operator_stack))

        for i in range(len(operator_stack) - 1, -1, -1):
            value2 = operand_stack.pop(i + 1)
            value1 = operand_stack.pop(i)
            operator = operator_stack.pop(i)

            if operator == "and":
                operand_stack.insert(i, value1 and value2)
            if operator == "or":
                operand_stack.insert(i, value1 or value2)
            if operator == "==":
                operand_stack.insert(i, value1 == value2)
            if operator == "!=":
                operand_stack.insert(i, value1 != value2)
            if operator == ">":
                operand_stack.insert(i, value1 > value2)
            if operator == ">=":
                operand_stack.insert(i, value1 >= value2)
            if operator == "<":
                operand_stack.insert(i, value1 < value2)
            if operator == "<=":
                operand_stack.insert(i, value1 <= value2)

        return operand_stack.pop()

    @staticmethod
    def parse(branch: List[AST]):
        return LogicExpression(branch)

    def __repr__(self):
        stack = []
        for a, b in zip(self.operand_stack, self.operator_stack):
            stack.append(a)
            stack.append(b)
        stack.append(self.operand_stack[-1])

        return " ".join(map(str, stack))



from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding


class Increment(AST[None]):
    def __init__(self, name: str):
        self.name = name

    def execute(self, binding: ASTBinding) -> None:
        value = binding.properties[self.name]
        binding.add_property(self.name, value + 1)
        return None

    @staticmethod
    def parse(name: str) -> 'Increment':
        return Increment(name)



from typing import Union, List

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding
from scriptable.ast.expression.logic_expression import LogicExpression

DataType = Union[int, float]


class LogicTerm(AST[DataType]):
    def __init__(self, branch: List[AST]):
        self.expression = LogicExpression(branch)

    def execute(self, binding: ASTBinding) -> DataType:
        return self.expression.execute(binding)

    @staticmethod
    def parse(branch: List[AST]):
        return LogicTerm(branch)



# Generated from Typescript.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .TypescriptParser import TypescriptParser
else:
    from TypescriptParser import TypescriptParser

# This class defines a complete generic visitor for a parse tree produced by TypescriptParser.

class TypescriptVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by TypescriptParser#sAll.
    def visitSAll(self, ctx:TypescriptParser.SAllContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sOperand.
    def visitSOperand(self, ctx:TypescriptParser.SOperandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sOperator.
    def visitSOperator(self, ctx:TypescriptParser.SOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sExpression.
    def visitSExpression(self, ctx:TypescriptParser.SExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sTerm.
    def visitSTerm(self, ctx:TypescriptParser.STermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sValue.
    def visitSValue(self, ctx:TypescriptParser.SValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sInvocation.
    def visitSInvocation(self, ctx:TypescriptParser.SInvocationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sControl.
    def visitSControl(self, ctx:TypescriptParser.SControlContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sStatement.
    def visitSStatement(self, ctx:TypescriptParser.SStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sPlus.
    def visitSPlus(self, ctx:TypescriptParser.SPlusContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sMinus.
    def visitSMinus(self, ctx:TypescriptParser.SMinusContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sMul.
    def visitSMul(self, ctx:TypescriptParser.SMulContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sDiv.
    def visitSDiv(self, ctx:TypescriptParser.SDivContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sPower.
    def visitSPower(self, ctx:TypescriptParser.SPowerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sAnd.
    def visitSAnd(self, ctx:TypescriptParser.SAndContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sOr.
    def visitSOr(self, ctx:TypescriptParser.SOrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sNot.
    def visitSNot(self, ctx:TypescriptParser.SNotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sEquals.
    def visitSEquals(self, ctx:TypescriptParser.SEqualsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sNotEquals.
    def visitSNotEquals(self, ctx:TypescriptParser.SNotEqualsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sLowerThan.
    def visitSLowerThan(self, ctx:TypescriptParser.SLowerThanContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sLowerEquals.
    def visitSLowerEquals(self, ctx:TypescriptParser.SLowerEqualsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sGreaterThan.
    def visitSGreaterThan(self, ctx:TypescriptParser.SGreaterThanContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sGreaterEquals.
    def visitSGreaterEquals(self, ctx:TypescriptParser.SGreaterEqualsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sArithmeticExpression.
    def visitSArithmeticExpression(self, ctx:TypescriptParser.SArithmeticExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sArithmeticTerm.
    def visitSArithmeticTerm(self, ctx:TypescriptParser.SArithmeticTermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sBooleanOperand.
    def visitSBooleanOperand(self, ctx:TypescriptParser.SBooleanOperandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sBooleanOperator.
    def visitSBooleanOperator(self, ctx:TypescriptParser.SBooleanOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sBooleanExpression.
    def visitSBooleanExpression(self, ctx:TypescriptParser.SBooleanExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sBooleanTerm.
    def visitSBooleanTerm(self, ctx:TypescriptParser.SBooleanTermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sNumberOperand.
    def visitSNumberOperand(self, ctx:TypescriptParser.SNumberOperandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sNumberOperator.
    def visitSNumberOperator(self, ctx:TypescriptParser.SNumberOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sNumberExpression.
    def visitSNumberExpression(self, ctx:TypescriptParser.SNumberExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sNumberTerm.
    def visitSNumberTerm(self, ctx:TypescriptParser.SNumberTermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sStringOperand.
    def visitSStringOperand(self, ctx:TypescriptParser.SStringOperandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sStringOperator.
    def visitSStringOperator(self, ctx:TypescriptParser.SStringOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sStringExpression.
    def visitSStringExpression(self, ctx:TypescriptParser.SStringExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sStringTerm.
    def visitSStringTerm(self, ctx:TypescriptParser.SStringTermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sConcatOperand.
    def visitSConcatOperand(self, ctx:TypescriptParser.SConcatOperandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sConcatExpression.
    def visitSConcatExpression(self, ctx:TypescriptParser.SConcatExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sConcatLeft.
    def visitSConcatLeft(self, ctx:TypescriptParser.SConcatLeftContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sConcatRight.
    def visitSConcatRight(self, ctx:TypescriptParser.SConcatRightContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sConcatBoth.
    def visitSConcatBoth(self, ctx:TypescriptParser.SConcatBothContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sType.
    def visitSType(self, ctx:TypescriptParser.STypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sFunction.
    def visitSFunction(self, ctx:TypescriptParser.SFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sFunctionArg.
    def visitSFunctionArg(self, ctx:TypescriptParser.SFunctionArgContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sFunctionArgs.
    def visitSFunctionArgs(self, ctx:TypescriptParser.SFunctionArgsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sFunctionArgDef.
    def visitSFunctionArgDef(self, ctx:TypescriptParser.SFunctionArgDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sFunctionArgDefs.
    def visitSFunctionArgDefs(self, ctx:TypescriptParser.SFunctionArgDefsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sFunctionHead.
    def visitSFunctionHead(self, ctx:TypescriptParser.SFunctionHeadContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sFunctionTail.
    def visitSFunctionTail(self, ctx:TypescriptParser.SFunctionTailContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sFunctionCall.
    def visitSFunctionCall(self, ctx:TypescriptParser.SFunctionCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sFunctionLambda.
    def visitSFunctionLambda(self, ctx:TypescriptParser.SFunctionLambdaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sProperty.
    def visitSProperty(self, ctx:TypescriptParser.SPropertyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sPropertyAware.
    def visitSPropertyAware(self, ctx:TypescriptParser.SPropertyAwareContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sPropertyAccess.
    def visitSPropertyAccess(self, ctx:TypescriptParser.SPropertyAccessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sFunctionAware.
    def visitSFunctionAware(self, ctx:TypescriptParser.SFunctionAwareContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sFunctionAccess.
    def visitSFunctionAccess(self, ctx:TypescriptParser.SFunctionAccessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sLine.
    def visitSLine(self, ctx:TypescriptParser.SLineContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sBody.
    def visitSBody(self, ctx:TypescriptParser.SBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sReturn.
    def visitSReturn(self, ctx:TypescriptParser.SReturnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sIf.
    def visitSIf(self, ctx:TypescriptParser.SIfContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sElse.
    def visitSElse(self, ctx:TypescriptParser.SElseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sElseIf.
    def visitSElseIf(self, ctx:TypescriptParser.SElseIfContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sString.
    def visitSString(self, ctx:TypescriptParser.SStringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sNumber.
    def visitSNumber(self, ctx:TypescriptParser.SNumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sBoolean.
    def visitSBoolean(self, ctx:TypescriptParser.SBooleanContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sArray.
    def visitSArray(self, ctx:TypescriptParser.SArrayContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sMap.
    def visitSMap(self, ctx:TypescriptParser.SMapContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sEndlessLoop.
    def visitSEndlessLoop(self, ctx:TypescriptParser.SEndlessLoopContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sWhile.
    def visitSWhile(self, ctx:TypescriptParser.SWhileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sFor.
    def visitSFor(self, ctx:TypescriptParser.SForContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sForOf.
    def visitSForOf(self, ctx:TypescriptParser.SForOfContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sForIn.
    def visitSForIn(self, ctx:TypescriptParser.SForInContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sLoopTail.
    def visitSLoopTail(self, ctx:TypescriptParser.SLoopTailContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sContinue.
    def visitSContinue(self, ctx:TypescriptParser.SContinueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sBreak.
    def visitSBreak(self, ctx:TypescriptParser.SBreakContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sMutableVar.
    def visitSMutableVar(self, ctx:TypescriptParser.SMutableVarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sImmutableVar.
    def visitSImmutableVar(self, ctx:TypescriptParser.SImmutableVarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sAssignment.
    def visitSAssignment(self, ctx:TypescriptParser.SAssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sIncrement.
    def visitSIncrement(self, ctx:TypescriptParser.SIncrementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by TypescriptParser#sDecrement.
    def visitSDecrement(self, ctx:TypescriptParser.SDecrementContext):
        return self.visitChildren(ctx)



del TypescriptParser


from .arithmetic_expression import ArithmeticExpression
from .arithmetic_term import ArithmeticTerm
from .concat_expression import ConcatExpression
from .logic_expression import LogicExpression
from .logic_term import LogicTerm
from .operator import And
from .operator import Div
from .operator import Equals
from .operator import GreaterEquals
from .operator import LowerEquals
from .operator import LowerEquals
from .operator import LowerThan
from .operator import Minus
from .operator import Mul
from .operator import Not
from .operator import NotEquals
from .operator import Or
from .operator import Plus
from .operator import Power



from typing import Any, List

from scriptable.api.accessor import Accessor


class TypescriptStringAccessor(Accessor[str]):

    def __init__(self, value: str):
        self.value = value

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.value[item]
        if item == "length":
            return len(self.value)
        raise ValueError(f"{item} property value access is not supported")

    def __call__(self, name: str, args: List[Any]):
        if name == "slice":
            return self._slice(args)
        if name == "substring":
            return self._substring(args)
        if name == "substr":
            return self._substr(args)
        if name == "replace":
            return self._replace(args)
        if name == "toUpperCase":
            return self.value.upper()
        if name == "toLowerCase":
            return self.value.lower()
        if name == "concat":
            return self._concat(args)
        if name == "trim":
            return self.value.strip()
        if name == "padStart":
            return self._pad_start(args)
        if name == "padEnd":
            return self._pad_end(args)
        if name == "charAt":
            return self._char_at(args)
        if name == "charCodeAt":
            return self._char_code_at(args)
        if name == "split":
            return self._split(args)
        raise ValueError(f"{name} function value access is not supported")

    def _slice(self, args: List[Any]):
        if len(args) == 1:
            if args[0] < 0:
                args[0] = len(self.value) + args[0]
            return self.value[args[0]:]
        else:
            if args[0] < 0:
                args[0] = len(self.value) + args[0]
            if args[1] < 0:
                args[1] = len(self.value) + args[1]
            return self.value[args[0]: args[1]]

    def _substring(self, args: List[Any]):
        if len(args) == 1:
            return self.value[args[0]:]
        return self.value[args[0]: args[1]]

    def _substr(self, args: List[Any]):
        if len(args) == 1:
            return self.value[args[0]:]
        return self.value[args[0]:args[0] + args[1]]

    def _replace(self, args: List[Any]):
        assert len(args) == 2, f"invalid arguments for replace {args}"
        assert isinstance(args[0], TypescriptStringAccessor), "argument is not a string"
        assert isinstance(args[1], TypescriptStringAccessor), "argument is not a string"
        return self.value.replace(args[0].value, args[1].value)

    def _concat(self, args: List[Any]):
        assert len(args) == 1, f"invalid arguments for concat {args}"
        assert isinstance(args[0], TypescriptStringAccessor), "argument is not a string"
        return TypescriptStringAccessor(self.value + args[0].value)

    def _pad_start(self, args: List[Any]):
        assert len(args) == 2, f"invalid arguments for replace {args}"
        assert isinstance(args[0], int), "argument is not an int"
        return self.value.rjust(args[0], str(args[1]))

    def _pad_end(self, args: List[Any]):
        assert len(args) == 2, f"invalid arguments for replace {args}"
        assert isinstance(args[0], int), "argument is not an int"
        return self.value.ljust(args[0], str(args[1]))

    def _char_at(self, args: List[Any]):
        assert len(args) == 1, f"invalid arguments for replace {args}"
        assert isinstance(args[0], int), "argument is not an int"
        return self.value[args[0]]

    def _char_code_at(self, args: List[Any]):
        assert len(args) == 1, f"invalid arguments for replace {args}"
        assert isinstance(args[0], int), "argument is not an int"
        return ord(self.value[args[0]])

    def _split(self, args: List[Any]):
        assert len(args) == 1, f"invalid arguments for replace {args}"
        assert isinstance(args[0], TypescriptStringAccessor), "argument is not a string"
        return self.value.split(args[0].value)

    def __len__(self):
        return len(self.value)

    def __repr__(self):
        return self.value


from typing import Union

from scriptable.antlr.TypescriptParser import TypescriptParser
from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding
from scriptable.api.ast_restrictions import ASTRestrictions

DataType = Union[int, float]


class Number(AST[DataType]):
    def __init__(self, value: DataType):
        self.value = value

    def execute(self, binding: ASTBinding) -> DataType:
        return self.value

    @staticmethod
    def parse(ctx: TypescriptParser.SNumberContext, settings: ASTRestrictions) -> 'Number':
        text = ctx.getText()
        assert len(text) <= settings.max_precision, f"max value precision exceeded"

        if "." in text:
            assert len(text[text.find("."):]) <= settings.max_scale, "max value scale exceeded"
            return Number(float(text))

        return Number(int(text))

    def __repr__(self):
        return str(self.value)



from typing import Any, Dict


class Process:

    def __init__(self, context: Dict[str, Any]):
        self.context = context

    def __call__(self, *args):
        pass

    def __getitem__(self, item):
        if item == "env":
            return self.context
        raise ValueError(f"cannot handle {item}")



from typing import List

from scriptable.api import AST
from scriptable.api.ast import EmptyAST
from scriptable.api.ast_binding import ASTBinding
from scriptable.api.exit_value import GoTo
from scriptable.ast.variable.assignment import Assignment


class For(AST[None]):

    def __init__(self, assignment: Assignment, expression: AST, statement: AST, tail: AST):
        self.assignment = assignment
        self.expression = expression
        self.statement = statement
        self.tail = tail

    def execute(self, context: ASTBinding) -> None:
        if self.assignment.name in context.properties:
            raise ValueError(f"property {self.assignment.name} is already defined")

        self.assignment.execute(context)
        value = context.properties[self.assignment.name]
        assert value <= context.restrictions.max_loops, "max loops exceeded"

        counter = 0
        while self.expression.execute(context):
            result = self.tail.execute(context)
            self.statement.execute(context)

            if isinstance(result, GoTo) and result.value == "break":
                break
            if isinstance(result, GoTo) and result.value == "continue":
                continue

            assert counter <= context.restrictions.max_loops, "max loops exceeded"
            counter = counter + 1

        return None

    @staticmethod
    def parse(branch: List[AST]) -> 'For':
        if len(branch) > 3:
            # noinspection PyTypeChecker
            return For(branch[0], branch[1], branch[2], branch[3])
        # noinspection PyTypeChecker
        return For(branch[0], branch[1], branch[2], EmptyAST())



from typing import Any, List, Tuple

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding


class FunctionArg(AST[Any]):
    def __init__(self, arg_value: AST[Any]):
        self.arg_value = arg_value

    def execute(self, binding: ASTBinding) -> Tuple[str, List[Any]]:
        return self.arg_value.execute(binding)

    @staticmethod
    def parse(branch: List[AST]) -> 'FunctionArg':
        return FunctionArg(branch[0])

    def __repr__(self):
        return str(self.arg_value)


from abc import ABC, abstractmethod
from typing import TypeVar, Generic

from scriptable.api.ast_binding import ASTBinding

T = TypeVar("T")


class AST(ABC, Generic[T]):

    @abstractmethod
    def execute(self, context: ASTBinding) -> T:
        pass


class EmptyAST(AST[None]):

    def execute(self, context: ASTBinding) -> None:
        return None



from dataclasses import dataclass
from typing import List

from antlr4.error.ErrorListener import ErrorListener


class ScriptableErrorListener(ErrorListener):
    errors: List['ScriptableError'] = []

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        if "no viable alternative at input" in msg:
            _msg = "compiler error: " + msg[len("no viable alternative at input"):]
            self.errors.append(ScriptableError(_msg, line, column))
        else:
            self.errors.append(ScriptableError(msg, line, column))

    def reportAmbiguity(self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs):
        pass

    def reportAttemptingFullContext(self, recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs):
        pass

    def reportContextSensitivity(self, recognizer, dfa, startIndex, stopIndex, prediction, configs):
        pass


@dataclass
class ScriptableError:
    msg: str
    line: int
    column: int

    def __repr__(self):
        return f"{self.msg} at {self.line}:{self.column}"



from typing import Union, List

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding

DataType = Union[int, float]


class ConcatExpression(AST[str]):
    def __init__(self, branch: List[AST]):
        self.operand_stack = [branch[x] for x in range(0, len(branch), 2)]
        self.operator_stack = [branch[x] for x in range(1, len(branch), 2)]

    def execute(self, binding: ASTBinding) -> str:
        from copy import deepcopy

        def execute_branch(ast: AST):
            result = ast
            while isinstance(result, AST):
                result = result.execute(deepcopy(binding))
            return result

        operand_stack = list(map(lambda ast: execute_branch(ast), self.operand_stack))
        operator_stack = list(map(lambda ast: execute_branch(ast), self.operator_stack))

        return "".join(map(str, operand_stack))

    @staticmethod
    def parse(branch: List[AST]):
        return ConcatExpression(branch)

    def __repr__(self):
        stack = []
        for a, b in zip(self.operand_stack, self.operator_stack):
            stack.append(a)
            stack.append(b)
        stack.append(self.operand_stack[-1])

        return " ".join(map(str, stack))



from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding


class And(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return "and"

    def __repr__(self):
        return "and"


class Or(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return "or"

    def __repr__(self):
        return "or"


class Not(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return "not"

    def __repr__(self):
        return "not"


class Plus(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return "+"

    def __repr__(self):
        return "+"


class Minus(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return "-"

    def __repr__(self):
        return "-"


class Mul(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return "*"

    def __repr__(self):
        return "*"


class Div(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return "/"

    def __repr__(self):
        return "/"


class Power(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return "**"

    def __repr__(self):
        return "**"


class Equals(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return "=="

    def __repr__(self):
        return "=="


class NotEquals(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return "!="

    def __repr__(self):
        return "!="


class LowerThan(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return "<"

    def __repr__(self):
        return "<"


class LowerEquals(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return "<="

    def __repr__(self):
        return "<="


class GreaterThan(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return ">"

    def __repr__(self):
        return ">"


class GreaterEquals(AST[str]):

    def execute(self, binding: ASTBinding) -> str:
        return ">="

    def __repr__(self):
        return ">="



from typing import Union, List

from scriptable.api import AST
from scriptable.api.accessor import Accessor
from scriptable.api.ast_binding import ASTBinding

DataType = Union[int, float]


class ArithmeticExpression(AST[DataType]):
    def __init__(self, branch: List[AST]):
        self.operand_stack = [branch[x] for x in range(0, len(branch), 2)]
        self.operator_stack = [branch[x] for x in range(1, len(branch), 2)]

    def execute(self, binding: ASTBinding) -> DataType:
        from copy import deepcopy

        def execute_branch(ast: AST):
            result = ast
            while isinstance(result, AST):
                result = result.execute(deepcopy(binding))
            return result

        def unwrap(obj):
            if isinstance(obj, Accessor):
                return obj.value
            return obj

        operand_stack = list(map(lambda ast: execute_branch(ast), self.operand_stack))
        operator_stack = list(map(lambda ast: execute_branch(ast), self.operator_stack))

        # search for exponent operator
        for i in range(len(operator_stack) - 1, -1, -1):
            if operator_stack[i] == "**":
                value2 = operand_stack.pop(i + 1)
                value1 = operand_stack.pop(i)
                operand_stack.insert(i, value1 ** value2)
                operator_stack.pop(i)

        # search for mul and div operator
        for i in range(len(operator_stack) - 1, -1, -1):
            if operator_stack[i] == "*":
                value2 = operand_stack.pop(i + 1)
                value1 = operand_stack.pop(i)
                operand_stack.insert(i, value1 * value2)
                operator_stack.pop(i)
            elif operator_stack[i] == "/":
                value2 = operand_stack.pop(i + 1)
                value1 = operand_stack.pop(i)
                operand_stack.insert(i, value1 / value2)
                operator_stack.pop(i)

        # search for plus and minus operator
        for i in range(len(operator_stack) - 1, -1, -1):
            if operator_stack[i] == "+":
                value2 = unwrap(operand_stack.pop(i + 1))
                value1 = unwrap(operand_stack.pop(i))
                operand_stack.insert(i, value1 + value2)
                operator_stack.pop(i)
            elif operator_stack[i] == "-":
                value2 = operand_stack.pop(i + 1)
                value1 = operand_stack.pop(i)
                operand_stack.insert(i, value1 - value2)
                operator_stack.pop(i)

        return operand_stack.pop()

    @staticmethod
    def parse(branch: List[AST]):
        return ArithmeticExpression(branch)

    def __repr__(self):
        stack = []
        for a, b in zip(self.operand_stack, self.operator_stack):
            stack.append(a)
            stack.append(b)
        stack.append(self.operand_stack[-1])

        return " ".join(map(str, stack))



from typing import List, Optional

from scriptable.api import AST
from scriptable.ast.function.function_arg_def import FunctionArgDef
from scriptable.ast.type import Type


class FunctionHead:
    def __init__(self, name: str, args: List[FunctionArgDef], r_type: Optional[AST]):
        self.name = name
        self.args = args
        self.r_type = r_type

    # noinspection PyTypeChecker
    @staticmethod
    def parse(name: str, branch: List[AST]) -> 'FunctionHead':
        if len(branch) > 0 and isinstance(branch[-1], Type):
            return FunctionHead(name, branch, branch[-1])
        return FunctionHead(name, branch, None)



from typing import Any, List, Tuple

from scriptable.api import AST
from scriptable.api.accessor import Accessor
from scriptable.api.ast_binding import ASTBinding, SourceAwareContext


class FunctionCall(AST[Tuple[str, List[Any]]]):
    def __init__(self, name: str, args: List[AST]):
        self.name = name
        self.args = args

    def execute(self, binding: ASTBinding) -> Tuple[str, List[Any]]:
        args = list(map(lambda x: x.execute(binding), self.args))

        if isinstance(binding, SourceAwareContext):
            source = binding.source
            if isinstance(source, Accessor):
                return source(self.name, args)
            if hasattr(source, self.name):
                return getattr(source, self.name)(*args)
            return source.__call__(*args)

        if self.name in binding.functions:
            from copy import deepcopy
            _binding = deepcopy(binding)

            # recursion guard
            # ***************
            signature = self.name + "(" + ", ".join(map(str, args)) + ")"
            assert _binding.add_signature(signature) < 2, "recursion loop determined"

            function = _binding.functions[self.name]
            return function(args, _binding)

        raise ValueError(f"cannot find function {self.name}")

    @staticmethod
    def parse(text: str, branch: List[AST]) -> 'FunctionCall':
        return FunctionCall(text, branch)

    def __repr__(self):
        return self.name + "(" + ", ".join(map(str, self.args)) + ")"



from dataclasses import dataclass, field
from typing import Dict, Callable, List, Any

from scriptable.api.ast_restrictions import ASTRestrictions


@dataclass
class ASTBinding:
    functions: Dict[str, Callable[[List[Any], 'ASTBinding'], Any]] = field(default_factory=lambda: {})
    properties: Dict[str, Any] = field(default_factory=lambda: {})
    signatures: Dict[str, int] = field(default_factory=lambda: {})
    restrictions: ASTRestrictions = ASTRestrictions()

    def add_property(self, name: str, value: Any):
        self.properties[name] = value

    def add_function(self, name: str, value: Callable[[List[Any]], Any]):
        self.functions[name] = value

    def add_signature(self, signature: str) -> int:
        if signature in self.signatures:
            self.signatures[signature] += 1
        else:
            self.signatures[signature] = 1
        return self.signatures[signature]


class SourceAwareContext(ASTBinding):
    def __init__(self, source: Any, context: ASTBinding):
        self.source = source
        self.functions = context.functions
        self.properties = context.properties



from typing import List, Any

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding
from scriptable.api.exit_value import ExitValue


class FunctionTail(AST[Any]):
    def __init__(self, branch: List[AST]):
        self.branch = branch

    def execute(self, context: ASTBinding) -> Any:
        result = None
        for branch in self.branch:
            result = branch.execute(context)
            if isinstance(result, ExitValue):
                return result.value

        return result

    @staticmethod
    def parse(branch: List[AST]) -> 'FunctionTail':
        return FunctionTail(branch)

    def __repr__(self):
        return " ".join(map(str, self.branch))



# Generated from Typescript.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .TypescriptParser import TypescriptParser
else:
    from TypescriptParser import TypescriptParser

# This class defines a complete listener for a parse tree produced by TypescriptParser.
class TypescriptListener(ParseTreeListener):

    # Enter a parse tree produced by TypescriptParser#sAll.
    def enterSAll(self, ctx:TypescriptParser.SAllContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sAll.
    def exitSAll(self, ctx:TypescriptParser.SAllContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sOperand.
    def enterSOperand(self, ctx:TypescriptParser.SOperandContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sOperand.
    def exitSOperand(self, ctx:TypescriptParser.SOperandContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sOperator.
    def enterSOperator(self, ctx:TypescriptParser.SOperatorContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sOperator.
    def exitSOperator(self, ctx:TypescriptParser.SOperatorContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sExpression.
    def enterSExpression(self, ctx:TypescriptParser.SExpressionContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sExpression.
    def exitSExpression(self, ctx:TypescriptParser.SExpressionContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sTerm.
    def enterSTerm(self, ctx:TypescriptParser.STermContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sTerm.
    def exitSTerm(self, ctx:TypescriptParser.STermContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sValue.
    def enterSValue(self, ctx:TypescriptParser.SValueContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sValue.
    def exitSValue(self, ctx:TypescriptParser.SValueContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sInvocation.
    def enterSInvocation(self, ctx:TypescriptParser.SInvocationContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sInvocation.
    def exitSInvocation(self, ctx:TypescriptParser.SInvocationContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sControl.
    def enterSControl(self, ctx:TypescriptParser.SControlContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sControl.
    def exitSControl(self, ctx:TypescriptParser.SControlContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sStatement.
    def enterSStatement(self, ctx:TypescriptParser.SStatementContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sStatement.
    def exitSStatement(self, ctx:TypescriptParser.SStatementContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sPlus.
    def enterSPlus(self, ctx:TypescriptParser.SPlusContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sPlus.
    def exitSPlus(self, ctx:TypescriptParser.SPlusContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sMinus.
    def enterSMinus(self, ctx:TypescriptParser.SMinusContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sMinus.
    def exitSMinus(self, ctx:TypescriptParser.SMinusContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sMul.
    def enterSMul(self, ctx:TypescriptParser.SMulContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sMul.
    def exitSMul(self, ctx:TypescriptParser.SMulContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sDiv.
    def enterSDiv(self, ctx:TypescriptParser.SDivContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sDiv.
    def exitSDiv(self, ctx:TypescriptParser.SDivContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sPower.
    def enterSPower(self, ctx:TypescriptParser.SPowerContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sPower.
    def exitSPower(self, ctx:TypescriptParser.SPowerContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sAnd.
    def enterSAnd(self, ctx:TypescriptParser.SAndContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sAnd.
    def exitSAnd(self, ctx:TypescriptParser.SAndContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sOr.
    def enterSOr(self, ctx:TypescriptParser.SOrContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sOr.
    def exitSOr(self, ctx:TypescriptParser.SOrContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sNot.
    def enterSNot(self, ctx:TypescriptParser.SNotContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sNot.
    def exitSNot(self, ctx:TypescriptParser.SNotContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sEquals.
    def enterSEquals(self, ctx:TypescriptParser.SEqualsContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sEquals.
    def exitSEquals(self, ctx:TypescriptParser.SEqualsContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sNotEquals.
    def enterSNotEquals(self, ctx:TypescriptParser.SNotEqualsContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sNotEquals.
    def exitSNotEquals(self, ctx:TypescriptParser.SNotEqualsContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sLowerThan.
    def enterSLowerThan(self, ctx:TypescriptParser.SLowerThanContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sLowerThan.
    def exitSLowerThan(self, ctx:TypescriptParser.SLowerThanContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sLowerEquals.
    def enterSLowerEquals(self, ctx:TypescriptParser.SLowerEqualsContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sLowerEquals.
    def exitSLowerEquals(self, ctx:TypescriptParser.SLowerEqualsContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sGreaterThan.
    def enterSGreaterThan(self, ctx:TypescriptParser.SGreaterThanContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sGreaterThan.
    def exitSGreaterThan(self, ctx:TypescriptParser.SGreaterThanContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sGreaterEquals.
    def enterSGreaterEquals(self, ctx:TypescriptParser.SGreaterEqualsContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sGreaterEquals.
    def exitSGreaterEquals(self, ctx:TypescriptParser.SGreaterEqualsContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sArithmeticExpression.
    def enterSArithmeticExpression(self, ctx:TypescriptParser.SArithmeticExpressionContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sArithmeticExpression.
    def exitSArithmeticExpression(self, ctx:TypescriptParser.SArithmeticExpressionContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sArithmeticTerm.
    def enterSArithmeticTerm(self, ctx:TypescriptParser.SArithmeticTermContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sArithmeticTerm.
    def exitSArithmeticTerm(self, ctx:TypescriptParser.SArithmeticTermContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sBooleanOperand.
    def enterSBooleanOperand(self, ctx:TypescriptParser.SBooleanOperandContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sBooleanOperand.
    def exitSBooleanOperand(self, ctx:TypescriptParser.SBooleanOperandContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sBooleanOperator.
    def enterSBooleanOperator(self, ctx:TypescriptParser.SBooleanOperatorContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sBooleanOperator.
    def exitSBooleanOperator(self, ctx:TypescriptParser.SBooleanOperatorContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sBooleanExpression.
    def enterSBooleanExpression(self, ctx:TypescriptParser.SBooleanExpressionContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sBooleanExpression.
    def exitSBooleanExpression(self, ctx:TypescriptParser.SBooleanExpressionContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sBooleanTerm.
    def enterSBooleanTerm(self, ctx:TypescriptParser.SBooleanTermContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sBooleanTerm.
    def exitSBooleanTerm(self, ctx:TypescriptParser.SBooleanTermContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sNumberOperand.
    def enterSNumberOperand(self, ctx:TypescriptParser.SNumberOperandContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sNumberOperand.
    def exitSNumberOperand(self, ctx:TypescriptParser.SNumberOperandContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sNumberOperator.
    def enterSNumberOperator(self, ctx:TypescriptParser.SNumberOperatorContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sNumberOperator.
    def exitSNumberOperator(self, ctx:TypescriptParser.SNumberOperatorContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sNumberExpression.
    def enterSNumberExpression(self, ctx:TypescriptParser.SNumberExpressionContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sNumberExpression.
    def exitSNumberExpression(self, ctx:TypescriptParser.SNumberExpressionContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sNumberTerm.
    def enterSNumberTerm(self, ctx:TypescriptParser.SNumberTermContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sNumberTerm.
    def exitSNumberTerm(self, ctx:TypescriptParser.SNumberTermContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sStringOperand.
    def enterSStringOperand(self, ctx:TypescriptParser.SStringOperandContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sStringOperand.
    def exitSStringOperand(self, ctx:TypescriptParser.SStringOperandContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sStringOperator.
    def enterSStringOperator(self, ctx:TypescriptParser.SStringOperatorContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sStringOperator.
    def exitSStringOperator(self, ctx:TypescriptParser.SStringOperatorContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sStringExpression.
    def enterSStringExpression(self, ctx:TypescriptParser.SStringExpressionContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sStringExpression.
    def exitSStringExpression(self, ctx:TypescriptParser.SStringExpressionContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sStringTerm.
    def enterSStringTerm(self, ctx:TypescriptParser.SStringTermContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sStringTerm.
    def exitSStringTerm(self, ctx:TypescriptParser.SStringTermContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sConcatOperand.
    def enterSConcatOperand(self, ctx:TypescriptParser.SConcatOperandContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sConcatOperand.
    def exitSConcatOperand(self, ctx:TypescriptParser.SConcatOperandContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sConcatExpression.
    def enterSConcatExpression(self, ctx:TypescriptParser.SConcatExpressionContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sConcatExpression.
    def exitSConcatExpression(self, ctx:TypescriptParser.SConcatExpressionContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sConcatLeft.
    def enterSConcatLeft(self, ctx:TypescriptParser.SConcatLeftContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sConcatLeft.
    def exitSConcatLeft(self, ctx:TypescriptParser.SConcatLeftContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sConcatRight.
    def enterSConcatRight(self, ctx:TypescriptParser.SConcatRightContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sConcatRight.
    def exitSConcatRight(self, ctx:TypescriptParser.SConcatRightContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sConcatBoth.
    def enterSConcatBoth(self, ctx:TypescriptParser.SConcatBothContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sConcatBoth.
    def exitSConcatBoth(self, ctx:TypescriptParser.SConcatBothContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sType.
    def enterSType(self, ctx:TypescriptParser.STypeContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sType.
    def exitSType(self, ctx:TypescriptParser.STypeContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sFunction.
    def enterSFunction(self, ctx:TypescriptParser.SFunctionContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sFunction.
    def exitSFunction(self, ctx:TypescriptParser.SFunctionContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sFunctionArg.
    def enterSFunctionArg(self, ctx:TypescriptParser.SFunctionArgContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sFunctionArg.
    def exitSFunctionArg(self, ctx:TypescriptParser.SFunctionArgContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sFunctionArgs.
    def enterSFunctionArgs(self, ctx:TypescriptParser.SFunctionArgsContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sFunctionArgs.
    def exitSFunctionArgs(self, ctx:TypescriptParser.SFunctionArgsContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sFunctionArgDef.
    def enterSFunctionArgDef(self, ctx:TypescriptParser.SFunctionArgDefContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sFunctionArgDef.
    def exitSFunctionArgDef(self, ctx:TypescriptParser.SFunctionArgDefContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sFunctionArgDefs.
    def enterSFunctionArgDefs(self, ctx:TypescriptParser.SFunctionArgDefsContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sFunctionArgDefs.
    def exitSFunctionArgDefs(self, ctx:TypescriptParser.SFunctionArgDefsContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sFunctionHead.
    def enterSFunctionHead(self, ctx:TypescriptParser.SFunctionHeadContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sFunctionHead.
    def exitSFunctionHead(self, ctx:TypescriptParser.SFunctionHeadContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sFunctionTail.
    def enterSFunctionTail(self, ctx:TypescriptParser.SFunctionTailContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sFunctionTail.
    def exitSFunctionTail(self, ctx:TypescriptParser.SFunctionTailContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sFunctionCall.
    def enterSFunctionCall(self, ctx:TypescriptParser.SFunctionCallContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sFunctionCall.
    def exitSFunctionCall(self, ctx:TypescriptParser.SFunctionCallContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sFunctionLambda.
    def enterSFunctionLambda(self, ctx:TypescriptParser.SFunctionLambdaContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sFunctionLambda.
    def exitSFunctionLambda(self, ctx:TypescriptParser.SFunctionLambdaContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sProperty.
    def enterSProperty(self, ctx:TypescriptParser.SPropertyContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sProperty.
    def exitSProperty(self, ctx:TypescriptParser.SPropertyContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sPropertyAware.
    def enterSPropertyAware(self, ctx:TypescriptParser.SPropertyAwareContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sPropertyAware.
    def exitSPropertyAware(self, ctx:TypescriptParser.SPropertyAwareContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sPropertyAccess.
    def enterSPropertyAccess(self, ctx:TypescriptParser.SPropertyAccessContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sPropertyAccess.
    def exitSPropertyAccess(self, ctx:TypescriptParser.SPropertyAccessContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sFunctionAware.
    def enterSFunctionAware(self, ctx:TypescriptParser.SFunctionAwareContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sFunctionAware.
    def exitSFunctionAware(self, ctx:TypescriptParser.SFunctionAwareContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sFunctionAccess.
    def enterSFunctionAccess(self, ctx:TypescriptParser.SFunctionAccessContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sFunctionAccess.
    def exitSFunctionAccess(self, ctx:TypescriptParser.SFunctionAccessContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sLine.
    def enterSLine(self, ctx:TypescriptParser.SLineContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sLine.
    def exitSLine(self, ctx:TypescriptParser.SLineContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sBody.
    def enterSBody(self, ctx:TypescriptParser.SBodyContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sBody.
    def exitSBody(self, ctx:TypescriptParser.SBodyContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sReturn.
    def enterSReturn(self, ctx:TypescriptParser.SReturnContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sReturn.
    def exitSReturn(self, ctx:TypescriptParser.SReturnContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sIf.
    def enterSIf(self, ctx:TypescriptParser.SIfContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sIf.
    def exitSIf(self, ctx:TypescriptParser.SIfContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sElse.
    def enterSElse(self, ctx:TypescriptParser.SElseContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sElse.
    def exitSElse(self, ctx:TypescriptParser.SElseContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sElseIf.
    def enterSElseIf(self, ctx:TypescriptParser.SElseIfContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sElseIf.
    def exitSElseIf(self, ctx:TypescriptParser.SElseIfContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sString.
    def enterSString(self, ctx:TypescriptParser.SStringContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sString.
    def exitSString(self, ctx:TypescriptParser.SStringContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sNumber.
    def enterSNumber(self, ctx:TypescriptParser.SNumberContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sNumber.
    def exitSNumber(self, ctx:TypescriptParser.SNumberContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sBoolean.
    def enterSBoolean(self, ctx:TypescriptParser.SBooleanContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sBoolean.
    def exitSBoolean(self, ctx:TypescriptParser.SBooleanContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sArray.
    def enterSArray(self, ctx:TypescriptParser.SArrayContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sArray.
    def exitSArray(self, ctx:TypescriptParser.SArrayContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sMap.
    def enterSMap(self, ctx:TypescriptParser.SMapContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sMap.
    def exitSMap(self, ctx:TypescriptParser.SMapContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sEndlessLoop.
    def enterSEndlessLoop(self, ctx:TypescriptParser.SEndlessLoopContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sEndlessLoop.
    def exitSEndlessLoop(self, ctx:TypescriptParser.SEndlessLoopContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sWhile.
    def enterSWhile(self, ctx:TypescriptParser.SWhileContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sWhile.
    def exitSWhile(self, ctx:TypescriptParser.SWhileContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sFor.
    def enterSFor(self, ctx:TypescriptParser.SForContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sFor.
    def exitSFor(self, ctx:TypescriptParser.SForContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sForOf.
    def enterSForOf(self, ctx:TypescriptParser.SForOfContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sForOf.
    def exitSForOf(self, ctx:TypescriptParser.SForOfContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sForIn.
    def enterSForIn(self, ctx:TypescriptParser.SForInContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sForIn.
    def exitSForIn(self, ctx:TypescriptParser.SForInContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sLoopTail.
    def enterSLoopTail(self, ctx:TypescriptParser.SLoopTailContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sLoopTail.
    def exitSLoopTail(self, ctx:TypescriptParser.SLoopTailContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sContinue.
    def enterSContinue(self, ctx:TypescriptParser.SContinueContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sContinue.
    def exitSContinue(self, ctx:TypescriptParser.SContinueContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sBreak.
    def enterSBreak(self, ctx:TypescriptParser.SBreakContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sBreak.
    def exitSBreak(self, ctx:TypescriptParser.SBreakContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sMutableVar.
    def enterSMutableVar(self, ctx:TypescriptParser.SMutableVarContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sMutableVar.
    def exitSMutableVar(self, ctx:TypescriptParser.SMutableVarContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sImmutableVar.
    def enterSImmutableVar(self, ctx:TypescriptParser.SImmutableVarContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sImmutableVar.
    def exitSImmutableVar(self, ctx:TypescriptParser.SImmutableVarContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sAssignment.
    def enterSAssignment(self, ctx:TypescriptParser.SAssignmentContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sAssignment.
    def exitSAssignment(self, ctx:TypescriptParser.SAssignmentContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sIncrement.
    def enterSIncrement(self, ctx:TypescriptParser.SIncrementContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sIncrement.
    def exitSIncrement(self, ctx:TypescriptParser.SIncrementContext):
        pass


    # Enter a parse tree produced by TypescriptParser#sDecrement.
    def enterSDecrement(self, ctx:TypescriptParser.SDecrementContext):
        pass

    # Exit a parse tree produced by TypescriptParser#sDecrement.
    def exitSDecrement(self, ctx:TypescriptParser.SDecrementContext):
        pass



del TypescriptParser


from typing import Any, List

from scriptable.api import AST
from scriptable.api.accessor import Accessor
from scriptable.api.ast_binding import ASTBinding
from scriptable.runtime.accessor.typescript.array_accessor import ArrayAccessor


class Array(AST[Accessor[List[Any]]]):
    def __init__(self, accessor: Accessor[str]):
        self.accessor = accessor

    def execute(self, binding: ASTBinding) -> Accessor[str]:
        return self.accessor

    @staticmethod
    def parse(branch: List[AST]) -> 'Array':
        return Array(ArrayAccessor(branch))



from typing import List, Dict, Any, Optional

# noinspection PyUnresolvedReferences
from antlr4 import InputStream, CommonTokenStream

from scriptable.antlr.TypescriptLexer import TypescriptLexer
from scriptable.antlr.TypescriptParser import TypescriptParser
from scriptable.api import AST
from scriptable.api.accessor import Accessor
from scriptable.api.ast_binding import ASTBinding
from scriptable.api.ast_restrictions import ASTRestrictions
from scriptable.api.exit_value import ExitValue
from scriptable.ast.number import Number
from scriptable.ast.value.array import Array
from scriptable.ast.value.map import Map
from scriptable.ast.value.string import String
from scriptable.listener.error_listener import ScriptableErrorListener
from scriptable.runtime.buildin.typescript.console import Console
from scriptable.runtime.buildin.typescript.process import Process
from scriptable.typescript_visitor import TypescriptVisitorImpl


class TypescriptEngine:

    @staticmethod
    def parse(schema, restrictions: ASTRestrictions = ASTRestrictions()) -> 'TypescriptEngine':
        lexer = TypescriptLexer(TypescriptEngine.stream(schema))
        stream = CommonTokenStream(lexer)
        parser = TypescriptParser(stream)

        error_listener = ScriptableErrorListener()
        parser.removeErrorListeners()
        parser.addErrorListener(error_listener)
        lexer.removeErrorListeners()
        lexer.addErrorListener(error_listener)

        visitor = TypescriptVisitorImpl(restrictions)
        tree = visitor.visit(parser.sAll())

        if len(error_listener.errors) > 0:
            raise ValueError(str(error_listener.errors[0]))

        return TypescriptEngine(tree, restrictions)

    def __init__(self, tree: List[AST], restrictions: ASTRestrictions):
        self.tree = tree
        self.restrictions = restrictions

    def execute(self, properties: Optional[dict] = None):
        properties = properties if properties is not None else {}

        def unwrap(obj):
            if isinstance(obj, List):
                return list(map(unwrap, obj))
            if isinstance(obj, Accessor):
                return unwrap(obj.value)
            if isinstance(obj, String):
                return obj.accessor.value
            if isinstance(obj, Number):
                return obj.value
            if isinstance(obj, Array):
                return unwrap(obj.accessor.value)
            if isinstance(obj, Map):
                return unwrap(obj.accessor.value)
            return obj

        binding = self._create_binding(properties, self.restrictions)
        result = None
        for branch in self.tree:
            result = branch.execute(binding)
            if isinstance(result, ExitValue):
                return unwrap(result.value)

        return unwrap(result)

    def _create_binding(self, properties: Dict[str, Any], restrictions: ASTRestrictions):
        binding = ASTBinding(restrictions=restrictions)
        binding.add_property("console", Console())
        binding.add_property("process", Process(properties))
        return binding

    @staticmethod
    def stream(obj):
        if isinstance(obj, InputStream):
            return obj
        if isinstance(obj, str):
            return InputStream(str(obj))
        if isinstance(obj, bytes):
            return InputStream(obj.decode("utf-8"))
        if isinstance(obj, bytearray):
            return InputStream(obj.decode("utf-8"))
        raise ValueError(str(obj) + " cannot be converted to input stream")



from scriptable.antlr.TypescriptParser import TypescriptParser
from scriptable.api import AST
from scriptable.api.accessor import Accessor
from scriptable.api.ast_binding import ASTBinding
from scriptable.runtime.accessor.typescript.string_accessor import TypescriptStringAccessor


class String(AST[Accessor[str]]):
    def __init__(self, accessor: Accessor[str]):
        self.accessor = accessor

    def execute(self, binding: ASTBinding) -> Accessor[str]:
        return self.accessor

    @staticmethod
    def parse(ctx: TypescriptParser.SStringContext) -> 'String':
        return String(TypescriptStringAccessor(ctx.getText()[1:-1]))



from typing import Any, List, Tuple, Optional

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding


class FunctionArgDef(AST[Any]):
    def __init__(self, arg_name: AST[Any], arg_type: Optional[AST[str]]):
        self.arg_name = arg_name
        self.arg_type = arg_type

    def execute(self, binding: ASTBinding) -> Tuple[str, List[Any]]:
        arg_name = self.arg_name.execute(binding)
        if self.arg_type is not None:
            arg_type = self.arg_type.execute(binding)
        return arg_name

    @staticmethod
    def parse(branch: List[AST]) -> 'FunctionArgDef':
        return FunctionArgDef(branch[0], None if len(branch) == 1 else branch[1])

    def __repr__(self):
        return str(self.arg_name) + "(" + str(self.arg_type) + ")"



from scriptable.antlr.TypescriptParser import TypescriptParser
from scriptable.antlr.TypescriptVisitor import TypescriptVisitor
from scriptable.api.ast_restrictions import ASTRestrictions
from scriptable.ast.Return import Return
from scriptable.ast.boolean import Boolean
from scriptable.ast.control.Break import Break
from scriptable.ast.control.Continue import Continue
from scriptable.ast.control.For import For
from scriptable.ast.control.ForIn import ForIn
from scriptable.ast.control.ForOf import ForOf
from scriptable.ast.control.While import While
from scriptable.ast.control.ifelse import If, Else, ElseIf
from scriptable.ast.control.loop_tail import LoopTail
from scriptable.ast.expression.arithmetic_expression import ArithmeticExpression
from scriptable.ast.expression.arithmetic_term import ArithmeticTerm
from scriptable.ast.expression.concat_expression import ConcatExpression
from scriptable.ast.expression.logic_expression import LogicExpression
from scriptable.ast.expression.logic_term import LogicTerm
from scriptable.ast.expression.operator import And, Or, Not, Plus, Minus, Mul, Div, Power, Equals, NotEquals, LowerThan
from scriptable.ast.expression.operator import LowerEquals, GreaterThan, GreaterEquals
from scriptable.ast.function.function import Function
from scriptable.ast.function.function_access import FunctionAccess
from scriptable.ast.function.function_arg import FunctionArg
from scriptable.ast.function.function_arg_def import FunctionArgDef
from scriptable.ast.function.function_call import FunctionCall
from scriptable.ast.function.function_head import FunctionHead
from scriptable.ast.function.function_lambda import FunctionLambda
from scriptable.ast.function.function_tail import FunctionTail
from scriptable.ast.number import Number
from scriptable.ast.property.property import Property
from scriptable.ast.property.property_access import PropertyAccess
from scriptable.ast.type import Type
from scriptable.ast.value.array import Array
from scriptable.ast.value.map import Map
from scriptable.ast.value.string import String
from scriptable.ast.variable.assignment import Assignment
from scriptable.ast.variable.decrement import Decrement
from scriptable.ast.variable.immutable_var import ImmutableVar
from scriptable.ast.variable.increment import Increment
from scriptable.ast.variable.mutable_var import MutableVar


class TypescriptVisitorImpl(TypescriptVisitor):

    def __init__(self, restrictions: ASTRestrictions):
        self.restrictions = restrictions

    def visitSNumber(self, ctx: TypescriptParser.SNumberContext):
        return Number.parse(ctx, self.restrictions)

    def visitSBoolean(self, ctx: TypescriptParser.SBooleanContext):
        return Boolean.parse(ctx)

    def visitSString(self, ctx: TypescriptParser.SStringContext):
        return String.parse(ctx)

    def visitSArray(self, ctx: TypescriptParser.SArrayContext):
        return Array.parse(super().visitSArray(ctx))

    def visitSMap(self, ctx: TypescriptParser.SMapContext):
        return Map.parse(super().visitSMap(ctx))

    def visitSArithmeticExpression(self, ctx: TypescriptParser.SArithmeticExpressionContext):
        return ArithmeticExpression.parse(super().visitSArithmeticExpression(ctx))

    def visitSArithmeticTerm(self, ctx: TypescriptParser.SArithmeticTermContext):
        return ArithmeticTerm.parse(super().visitSArithmeticTerm(ctx))

    def visitSBooleanTerm(self, ctx: TypescriptParser.SBooleanTermContext):
        return LogicTerm.parse(super().visitSBooleanTerm(ctx))

    def visitSBooleanExpression(self, ctx: TypescriptParser.SBooleanExpressionContext):
        return LogicExpression.parse(super().visitSBooleanExpression(ctx))

    def visitSNumberExpression(self, ctx: TypescriptParser.SNumberExpressionContext):
        return LogicExpression.parse(super().visitSNumberExpression(ctx))

    def visitSNumberTerm(self, ctx: TypescriptParser.SNumberTermContext):
        return LogicTerm.parse(super().visitSNumberTerm(ctx))

    def visitSStringExpression(self, ctx: TypescriptParser.SStringExpressionContext):
        return LogicExpression.parse(super().visitSStringExpression(ctx))

    def visitSConcatExpression(self, ctx: TypescriptParser.SConcatExpressionContext):
        return ConcatExpression.parse(super().visitSConcatExpression(ctx))

    def visitSStringTerm(self, ctx: TypescriptParser.SStringTermContext):
        return LogicTerm.parse(super().visitSStringTerm(ctx))

    def visitSAnd(self, ctx: TypescriptParser.SAndContext):
        return And()

    def visitSOr(self, ctx: TypescriptParser.SOrContext):
        return Or()

    def visitSNot(self, ctx: TypescriptParser.SNotContext):
        return Not()

    def visitSPlus(self, ctx: TypescriptParser.SPlusContext):
        return Plus()

    def visitSMinus(self, ctx: TypescriptParser.SMinusContext):
        return Minus()

    def visitSMul(self, ctx: TypescriptParser.SMulContext):
        return Mul()

    def visitSDiv(self, ctx: TypescriptParser.SDivContext):
        return Div()

    def visitSPower(self, ctx: TypescriptParser.SPowerContext):
        return Power()

    def visitSEquals(self, ctx: TypescriptParser.SEqualsContext):
        return Equals()

    def visitSNotEquals(self, ctx: TypescriptParser.SNotEqualsContext):
        return NotEquals()

    def visitSLowerThan(self, ctx: TypescriptParser.SLowerThanContext):
        return LowerThan()

    def visitSLowerEquals(self, ctx: TypescriptParser.SLowerEqualsContext):
        return LowerEquals()

    def visitSGreaterThan(self, ctx: TypescriptParser.SGreaterThanContext):
        return GreaterThan()

    def visitSGreaterEquals(self, ctx: TypescriptParser.SGreaterEqualsContext):
        return GreaterEquals()

    def visitSProperty(self, ctx: TypescriptParser.SPropertyContext):
        return Property.parse(ctx)

    def visitSPropertyAccess(self, ctx: TypescriptParser.SPropertyAccessContext):
        return PropertyAccess.parse(super().visitSPropertyAccess(ctx))

    def visitSFunctionCall(self, ctx: TypescriptParser.SFunctionCallContext):
        return FunctionCall.parse(ctx.getChild(0).getText(), super().visitSFunctionCall(ctx))

    def visitSFunctionAccess(self, ctx: TypescriptParser.SFunctionAccessContext):
        return FunctionAccess(super().visitSFunctionAccess(ctx))

    def visitSReturn(self, ctx: TypescriptParser.SReturnContext):
        return Return.parse(super().visitSReturn(ctx))

    def visitSIf(self, ctx: TypescriptParser.SIfContext):
        return If.parse(super().visitSIf(ctx))

    def visitSElse(self, ctx: TypescriptParser.SElseContext):
        return Else.parse(super().visitSElse(ctx))

    def visitSElseIf(self, ctx: TypescriptParser.SElseIfContext):
        return ElseIf.parse(super().visitSElseIf(ctx))

    def visitSType(self, ctx: TypescriptParser.STypeContext):
        return Type.parse(ctx)

    def visitSFunctionArg(self, ctx: TypescriptParser.SFunctionArgContext):
        return FunctionArg.parse(super().visitSFunctionArg(ctx))

    def visitSFunctionArgDef(self, ctx: TypescriptParser.SFunctionArgDefContext):
        return FunctionArgDef.parse(super().visitSFunctionArgDef(ctx))

    def visitSFunctionHead(self, ctx: TypescriptParser.SFunctionHeadContext):
        return FunctionHead.parse(ctx.getChild(1).getText(), super().visitSFunctionHead(ctx))

    def visitSFunctionTail(self, ctx: TypescriptParser.SFunctionTailContext):
        return FunctionTail.parse(super().visitSFunctionTail(ctx))

    def visitSFunction(self, ctx: TypescriptParser.SFunctionContext):
        return Function.parse(super().visitSFunction(ctx))

    def visitSFunctionLambda(self, ctx: TypescriptParser.SFunctionLambdaContext):
        return FunctionLambda.parse(super().visitSFunctionLambda(ctx))

    def visitSWhile(self, ctx: TypescriptParser.SWhileContext):
        return While.parse(super().visitSWhile(ctx))

    def visitSMutableVar(self, ctx: TypescriptParser.SMutableVarContext):
        return MutableVar.parse(ctx.getChild(1).getText(), super().visitSMutableVar(ctx)[0])

    def visitSImmutableVar(self, ctx: TypescriptParser.SImmutableVarContext):
        return ImmutableVar.parse(ctx.getChild(1).getText(), super().visitSImmutableVar(ctx)[0])

    def visitSAssignment(self, ctx: TypescriptParser.SAssignmentContext):
        return Assignment.parse(ctx.getChild(0).getText(), super().visitSAssignment(ctx)[0])

    def visitSEndlessLoop(self, ctx: TypescriptParser.SEndlessLoopContext):
        raise ValueError("endless loops are not supported")

    def visitSFor(self, ctx: TypescriptParser.SForContext):
        return For.parse(super().visitSFor(ctx))

    def visitSForOf(self, ctx: TypescriptParser.SForOfContext):
        return ForOf.parse(ctx.getChild(3).getText(), super().visitSForOf(ctx))

    def visitSForIn(self, ctx: TypescriptParser.SForInContext):
        return ForIn.parse(ctx.getChild(3).getText(), super().visitSForIn(ctx))

    def visitSIncrement(self, ctx: TypescriptParser.SIncrementContext):
        return Increment.parse(ctx.getChild(0).getText())

    def visitSDecrement(self, ctx: TypescriptParser.SDecrementContext):
        return Decrement.parse(ctx.getChild(0).getText())

    def visitSLoopTail(self, ctx: TypescriptParser.SLoopTailContext):
        return LoopTail.parse(super().visitSLoopTail(ctx))

    def visitSContinue(self, ctx: TypescriptParser.SContinueContext):
        return Continue()

    def visitSBreak(self, ctx: TypescriptParser.SBreakContext):
        return Break()

    def aggregateResult(self, aggregate, next_result):
        array = []
        if aggregate is not None:
            array += aggregate
        if next_result is not None:
            if isinstance(next_result, list):
                array.extend(next_result)
            else:
                array += [next_result]
        return array



from typing import List

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding
from scriptable.ast.function.function_head import FunctionHead
from scriptable.ast.function.function_tail import FunctionTail


class Function(AST[None]):
    def __init__(self, head: FunctionHead, tail: FunctionTail):
        self.head = head
        self.tail = tail

    def execute(self, context: ASTBinding) -> None:
        def execute_function(args, binding):
            from copy import deepcopy
            _binding = deepcopy(binding)

            for i, e in enumerate(self.head.args):
                arg_name = e.arg_name.value
                _binding.add_property(arg_name, args[i])

            return self.tail.execute(_binding)

        context.add_function(self.head.name, execute_function)
        return None

    @staticmethod
    def parse(branch: List[AST]) -> 'Function':
        # noinspection PyTypeChecker
        return Function(branch[0], branch[1])

    def __repr__(self):
        return str(self.head) + str(self.tail)



from typing import List

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding
from scriptable.api.exit_value import GoTo


class LoopTail(AST[GoTo]):

    def __init__(self, branch: List[AST]):
        self.branch = branch

    def execute(self, context: ASTBinding) -> GoTo:
        for item in self.branch:
            result = item.execute(context)
            if isinstance(result, GoTo):
                return result
        return GoTo("end-of-loop")

    @staticmethod
    def parse(branch: List[AST]) -> 'LoopTail':
        return LoopTail(branch)



from typing import Any, List, Dict

from scriptable.api.accessor import Accessor


class MapAccessor(Accessor[Dict[str, Any]]):

    def __init__(self, value: Dict[str, Any]):
        self.value = value

    def __getitem__(self, item):
        raise ValueError(f"{item} property value access is not supported")

    def __call__(self, name: str, args: List[Any]):
        raise ValueError(f"{name} function value access is not supported")

    def __len__(self):
        return len(self.value)

    def __repr__(self):
        return ", ".join(map(str, self.value))



from typing import Any, List, Optional

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding
from scriptable.api.exit_value import ExitValue, GoTo


class If(AST[Optional[Any]]):

    def __init__(self, if_branch: List[AST], else_if_branch: List[AST], else_branch: List[AST]):
        self.if_branch = if_branch
        self.else_if_branch = else_if_branch
        self.else_branch = else_branch

    def execute(self, binding: ASTBinding) -> Optional[Any]:
        expression = self.if_branch[0].execute(binding)
        if expression:
            for branch in self.if_branch[1:]:
                result = branch.execute(binding)
                if isinstance(result, ExitValue):
                    return result
                if isinstance(result, GoTo):
                    return result

        for branch in self.else_if_branch:
            result = branch.execute(binding)
            if isinstance(result, ExitValue):
                return result
            if isinstance(result, GoTo):
                return result

        for branch in self.else_branch:
            result = branch.execute(binding)
            if isinstance(result, ExitValue):
                return result
            if isinstance(result, GoTo):
                return result

        return None

    @staticmethod
    def parse(branch: List[AST]) -> 'If':
        if_branch = list(filter(lambda x: not isinstance(x, ElseIf) and not isinstance(x, Else), branch))
        else_if_branch = list(filter(lambda x: isinstance(x, ElseIf), branch))
        else_branch = list(filter(lambda x: isinstance(x, Else), branch))

        return If(if_branch, else_if_branch, else_branch)


class ElseIf(AST[Optional[Any]]):

    def __init__(self, branch: List[AST]):
        self.branch = branch

    def execute(self, binding: ASTBinding) -> Optional[Any]:
        expression = self.branch[0].execute(binding)
        if expression:
            for branch in self.branch[1:]:
                result = branch.execute(binding)
                if isinstance(result, ExitValue):
                    return result
                if isinstance(result, GoTo):
                    return result
        return None

    @staticmethod
    def parse(branch: List[AST]) -> 'ElseIf':
        return ElseIf(branch)


class Else(AST[Optional[Any]]):

    def __init__(self, branch: List[AST]):
        self.branch = branch

    def execute(self, binding: ASTBinding) -> Optional[Any]:
        for branch in self.branch:
            result = branch.execute(binding)
            if isinstance(result, ExitValue):
                return result
            if isinstance(result, GoTo):
                return result
        return None

    @staticmethod
    def parse(branch: List[AST]) -> 'Else':
        return Else(branch)



from typing import Any

from scriptable.antlr.TypescriptParser import TypescriptParser
from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding


class Property(AST[Any]):
    def __init__(self, value: str):
        self.value = value

    def execute(self, binding: ASTBinding) -> Any:
        if self.value in binding.properties:
            return binding.properties[self.value]
        return self.value

    @staticmethod
    def parse(ctx: TypescriptParser.SPropertyContext) -> 'Property':
        return Property(ctx.getText())

    def __repr__(self):
        return self.value



from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding
from scriptable.api.exit_value import GoTo


class Break(AST[GoTo]):

    def execute(self, context: ASTBinding) -> GoTo:
        return GoTo("break")

    @staticmethod
    def parse() -> 'Break':
        return Break()



from typing import List, Any

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding
from scriptable.ast.function.function_arg_def import FunctionArgDef


class FunctionLambda(AST[Any]):
    def __init__(self, head: List[AST], tail: List[AST]):
        self.head = head
        self.tail = tail

    def execute(self, context: ASTBinding) -> Any:
        def callback(args):
            from copy import deepcopy
            _binding = deepcopy(context)

            for i, e in enumerate(self.head):
                arg_name = e.arg_name.value
                _binding.add_property(arg_name, args[i])

            return self.tail[0].execute(_binding)

        return callback

    @staticmethod
    def parse(branch: List[AST]) -> 'FunctionLambda':
        head = list(filter(lambda x: isinstance(x, FunctionArgDef), branch))
        tail = list(filter(lambda x: not isinstance(x, FunctionArgDef), branch))
        return FunctionLambda(head, tail)

    def __repr__(self):
        return str(self.head) + str(self.tail)



from dataclasses import dataclass


@dataclass
class ASTRestrictions:
    max_precision = 10  # The max allowed number of digits in a number.
    max_scale = 4  # The max allowed digits to the right of the decimal point in a number.
    max_loops = 10 # The max allowed number of loop iterations.



import unittest

from scriptable.typescript_engine import TypescriptEngine


class TypescriptEngineTest(unittest.TestCase):

    def test_max_precision(self):
        with self.assertRaises(AssertionError) as error:
            TypescriptEngine.parse("10000000000 + 2")
        self.assertEqual("max value precision exceeded", error.exception.args[0])

    def test_max_scale(self):
        with self.assertRaises(AssertionError) as error:
            TypescriptEngine.parse("1.00000 + 2")
        self.assertEqual("max value scale exceeded", error.exception.args[0])

    def test_expression(self):
        self.assertAlmostEqual(TypescriptEngine.parse("1 + 2").execute(), 3)
        self.assertAlmostEqual(TypescriptEngine.parse("1.1 + 2.2").execute(), 3.3)

        self.assertAlmostEqual(TypescriptEngine.parse("1 - 2").execute(), -1)
        self.assertAlmostEqual(TypescriptEngine.parse("1 * 2").execute(), 2)
        self.assertAlmostEqual(TypescriptEngine.parse("1 / 2").execute(), 0.5)

    def test_arithmetic_order(self):
        self.assertAlmostEqual(TypescriptEngine.parse("1 + 2 * 3").execute(), 7)
        self.assertAlmostEqual(TypescriptEngine.parse("1 + 2 * 3 - 1 / 2").execute(), 6.5)
        self.assertAlmostEqual(TypescriptEngine.parse("(1 + 2) * 3").execute(), 9)
        self.assertAlmostEqual(TypescriptEngine.parse("(1 + 2) * (3 - 1) / 2").execute(), 3)

    def test_logic_expression(self):
        # boolean expression
        self.assertEqual(TypescriptEngine.parse("true && true").execute(), True)
        self.assertEqual(TypescriptEngine.parse("true && false").execute(), False)
        self.assertEqual(TypescriptEngine.parse("false && true").execute(), False)
        self.assertEqual(TypescriptEngine.parse("false && false").execute(), False)
        self.assertEqual(TypescriptEngine.parse("true || true").execute(), True)
        self.assertEqual(TypescriptEngine.parse("true || false").execute(), True)
        self.assertEqual(TypescriptEngine.parse("false || true").execute(), True)
        self.assertEqual(TypescriptEngine.parse("false || false").execute(), False)
        self.assertEqual(TypescriptEngine.parse("true && true || false").execute(), True)
        self.assertEqual(TypescriptEngine.parse("true && false || false").execute(), False)
        self.assertEqual(TypescriptEngine.parse("true == true").execute(), True)
        self.assertEqual(TypescriptEngine.parse("true === true").execute(), True)
        self.assertEqual(TypescriptEngine.parse("true == false").execute(), False)
        self.assertEqual(TypescriptEngine.parse("true === false").execute(), False)

        # number expression
        self.assertEqual(TypescriptEngine.parse("1 == 1").execute(), True)
        self.assertEqual(TypescriptEngine.parse("1 != 1").execute(), False)
        self.assertEqual(TypescriptEngine.parse("1 > 1").execute(), False)
        self.assertEqual(TypescriptEngine.parse("1 >= 1").execute(), True)
        self.assertEqual(TypescriptEngine.parse("1 < 1").execute(), False)
        self.assertEqual(TypescriptEngine.parse("1 <= 1").execute(), True)

        # string expression
        self.assertEqual(TypescriptEngine.parse("'test' == 'test'").execute(), True)
        self.assertEqual(TypescriptEngine.parse("'test' != 'test'").execute(), False)
        self.assertEqual(TypescriptEngine.parse("'a' + 'b' + 'c'").execute(), "abc")

    def test_parenthesizes(self):
        self.assertAlmostEqual(TypescriptEngine.parse("(((1 + 2)))").execute(), 3)
        self.assertAlmostEqual(TypescriptEngine.parse("(((true && true)))").execute(), True)
        self.assertAlmostEqual(TypescriptEngine.parse("false && (false || true)").execute(), False)

    def test_property_access(self):
        # self.assertAlmostEqual(TypescriptEngine.parse("'test'.length").execute(), 4)
        self.assertEqual(TypescriptEngine.parse("'abcd'[1]").execute(), "b")

    def test_function_access(self):
        self.assertEqual(TypescriptEngine.parse("'abc'.slice(-1)").execute(), "c")
        self.assertEqual(TypescriptEngine.parse("'abc'.substring(1)").execute(), "bc")
        self.assertEqual(TypescriptEngine.parse("'abc'.substring(1, 2)").execute(), "b")
        self.assertEqual(TypescriptEngine.parse("'abcdefg'.substr(2, 3)").execute(), "cde")
        self.assertEqual(TypescriptEngine.parse("'abcdefg'.replace('cde', '')").execute(), "abfg")
        self.assertEqual(TypescriptEngine.parse("'abc'.concat('defg')").execute(), "abcdefg")
        self.assertEqual(TypescriptEngine.parse("'  abc  '.trim()").execute(), "abc")
        self.assertEqual(TypescriptEngine.parse("'abc'.padStart(5, 0)").execute(), "00abc")
        self.assertEqual(TypescriptEngine.parse("'abc'.padEnd(5, 0)").execute(), "abc00")
        self.assertEqual(TypescriptEngine.parse("'abc'.charAt(1)").execute(), "b")
        self.assertEqual(TypescriptEngine.parse("'abcd'.charCodeAt(3)").execute(), 100)
        self.assertEqual(TypescriptEngine.parse("'a,b,c'.split(',')").execute(), ["a", "b", "c"])

    def test_if_control(self):
        self.assertEqual(TypescriptEngine.parse("if (true) { return 1 } return 0").execute(), 1)
        self.assertEqual(TypescriptEngine.parse("if (false) { return 1 } return 0").execute(), 0)
        self.assertEqual(TypescriptEngine.parse("if (false) {return 0} else if(true) {return 1} return 2").execute(), 1)
        self.assertEqual(TypescriptEngine.parse("if (false) {return 0} else {return 1} return 2").execute(), 1)

    def test_console(self):
        TypescriptEngine.parse("console.assert('a')").execute()

    def test_value(self):
        # array
        self.assertEqual(TypescriptEngine.parse("['a', 'b', 'c']").execute(), ["a", "b", "c"])
        self.assertEqual(TypescriptEngine.parse("[1, 2, 3]").execute(), [1, 2, 3])
        self.assertEqual(TypescriptEngine.parse("[[1, 2, 3]]").execute(), [[1, 2, 3]])
        # map
        self.assertEqual(TypescriptEngine.parse("{'a':0, 'b':1, 'c':2}").execute(), {"a": 0, "b": 1, "c": 2})
        # string
        self.assertEqual(TypescriptEngine.parse("'abc'").execute(), "abc")

    def test_function_without_args(self):
        engine = TypescriptEngine.parse("""
function test() {
   return "test"
}
test()
        """)
        self.assertEqual(engine.execute(), "test")

    def test_function_with_args(self):
        engine = TypescriptEngine.parse("""
function test(a:string, b:string) {
   return a + b
}
test('te', 'st')
        """)
        self.assertEqual(engine.execute(), "test")

    def test_overloading(self):
        self.assertEqual(TypescriptEngine.parse("'a' + 'b' + 'c'").execute(), "abc")
        self.assertEqual(TypescriptEngine.parse("function test(a) { return a } test('a' + 'b' + 'c')").execute(), "abc")

    def test_fibonacci(self):
        engine = TypescriptEngine.parse("""
function fibonacci(n:number) {
   if (n == 0) return 0
   if (n == 1) return 1
   if (n == 2) return 1
   return fibonacci(n - 1) + fibonacci(n - 2)
}
fibonacci(process.env.n)
        """)
        self.assertEqual(engine.execute({"n": 8}), 21)

    def test_recursion_guard(self):
        with self.assertRaises(AssertionError) as error:
            TypescriptEngine.parse("""
    function test(n:number) {
       return test(n)
    }
    test(8)
            """).execute()
            self.assertEqual(error.exception.args[0], "recursion loop determined")

    def test_lambda(self):
        self.assertEqual(TypescriptEngine.parse("[4, 3, 2, 1].sort((a, b) => a - b)").execute(), [1, 2, 3, 4])

    def test_mutable_variable(self):
        engine = TypescriptEngine.parse("""
let test = 5
test
        """)
        self.assertEqual(engine.execute(), 5)

    def test_immutable_variable(self):
        engine = TypescriptEngine.parse("""
const test = 5
test
        """)
        self.assertEqual(engine.execute(), 5)

    def test_while(self):
        engine = TypescriptEngine.parse("""
let n = 5
while (n > 0) {
    n = n - 1
}
n
        """)
        self.assertEqual(engine.execute(), 0)

    def test_for(self):
        engine = TypescriptEngine.parse("""
let a = 5
for (let n = 5; n > 0; n--) {
    a = a - 1
}
a
        """)
        self.assertEqual(engine.execute(), 0)

    def test_for_of(self):
        engine = TypescriptEngine.parse("""
let a = ""
for (let char of "test") {
    a = a + char
}
a
        """)
        self.assertEqual(engine.execute(), "test")

    def test_for_in(self):
        engine = TypescriptEngine.parse("""
let a = []
for (let i in [1, 2, 3, 4]) {
    a.push(i)
}
a
        """)
        self.assertEqual(engine.execute(), [0, 1, 2, 3])

    def test_loop_guard(self):
        with self.assertRaises(AssertionError) as error:
            TypescriptEngine.parse("for (let i = 12; i > 0; i--) {}").execute()
        self.assertEqual(error.exception.args[0], "max loops exceeded")

    def test_break(self):
        engine = TypescriptEngine.parse("""
let a = 0
for (let n = 5; n > 0; n--) {
    if (n == 2) break
    a = a + 1
}
a
            """)
        self.assertEqual(engine.execute(), 3)

    def test_continue(self):
        engine = TypescriptEngine.parse("""
let a = 0
for (let n = 5; n > 0; n--) {
    if (n == 2) continue
    a = a + 1
}
a
            """)
        self.assertEqual(engine.execute(), 4)

    def test_process_env(self):
        self.assertEqual(TypescriptEngine.parse("process.env.test").execute({"test": "abc"}), "abc")
        self.assertEqual(TypescriptEngine.parse("process.env['test']").execute({"test": "abc"}), "abc")



from setuptools import find_packages, setup

# read the contents of your README file
from pathlib import Path
long_description = Path(__file__).with_name("README.md").read_text()

setup(
    name='scriptable',
    packages=find_packages(exclude=("__tests__",)),
    version='0.1.1',
    license='Apache Software License',
    description='Scriptable is a sand-boxed scripting engine which can be used safely in an embedded environment.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='x-and-y',
    author_email='christian.weber@leftshift.one',
    url='https://github.com/c7nw3r/scriptable',
    download_url='https://github.com/c7nw3r/scriptable/archive/refs/tags/v0.1.1.tar.gz',
    keywords=['scripting-engine'],
    setup_requires=['setuptools_scm'],
    include_package_data=True,
    install_requires=[
        'antlr4-python3-runtime'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Interpreters',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



from typing import Any, List

from scriptable.api.accessor import Accessor


class ArrayAccessor(Accessor[List[Any]]):

    def __init__(self, value: List[Any]):
        self.value = value

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.value[item]
        raise ValueError(f"{item} property value access is not supported")

    def __call__(self, name: str, args: List[Any]):
        if name == "sort":
            return self._sort(args)
        if name == "push":
            self.value.append(args[0])
            return None

        raise ValueError(f"{name} function value access is not supported")

    def _sort(self, args: List[Any]):
        def compare(a, b):
            return args[0]([a, b])
        from functools import cmp_to_key
        return sorted(self.value, key=cmp_to_key(compare))

    def __len__(self):
        return len(self.value)

    def __repr__(self):
        return ", ".join(map(str, self.value))



from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Any, List

T = TypeVar("T")


class Accessor(ABC, Generic[T]):
    value: T

    @abstractmethod
    def __getitem__(self, item):
        pass

    @abstractmethod
    def __call__(self, name: str, args: List[Any]):
        pass

    def __eq__(self, other):
        if isinstance(other, Accessor):
            return self.value == other.value
        return self.value == other



import logging


class Console:

    def log(self, *args):
        """
        Outputs a message to the console
        :param args:
        """
        logging.warning(" ".join(map(str, args)))

    # FIXME: function name
    def _assert(self, *args):
        """
        Writes an error message to the console if a assertion is false
        :param args:
        """
        if args[0] == True:
            logging.error(" ".join(map(str, args[1:])))

    def clear(self):
        """
        Clears the console.
        """
        raise ValueError("not supported")

    def count(self):
        """
        Logs the number of times that this particular call to count() has been called
        """
        raise ValueError("not supported")

    def error(self, *args):
        """
        Outputs an error message to the console
        :param args:
        """
        logging.error(" ".join(map(str, args)))

    def group(self):
        """
        Creates a new inline group in the console. This indents following console messages by an additional level,
        until console.groupEnd() is called
        """
        raise ValueError("not supported")

    def group_collapsed(self):
        """
        Creates a new inline group in the console. However, the new group is created collapsed. The user will need to use the disclosure button to expand it
        """
        raise ValueError("not supported")

    def group_end(self):
        """
        Exits the current inline group in the console
        """
        raise ValueError("not supported")

    def info(self, *args):
        """
        Outputs an informational message to the console
        :param args:
        """
        logging.info(" ".join(map(str, args)))

    def table(self):
        """
        Displays tabular data as a table
        """
        raise ValueError("not supported")

    def time(self):
        """
        Starts a timer (can track how long an operation takes)
        """
        raise ValueError("not supported")

    def time_end(self):
        """
        Stops a timer that was previously started by console.time()
        :return:
        """
        raise ValueError("not supported")

    def trace(self, *args):
        """
        Outputs a stack trace to the console
        :param args:
        """
        logging.info(" ".join(map(str, args)))

    def warn(self, *args):
        """
        Outputs a warning message to the console
        :param args:
        """
        logging.warning(" ".join(map(str, args)))

    def __call__(self, *args):
        pass


from scriptable.antlr.TypescriptParser import TypescriptParser
from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding


class Boolean(AST[bool]):
    def __init__(self, value: bool):
        self.value = value

    def execute(self, binding: ASTBinding) -> bool:
        return self.value

    @staticmethod
    def parse(ctx: TypescriptParser.SBooleanContext) -> 'Boolean':
        text = ctx.getText()
        return Boolean(text == "true")



from typing import Any, List

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding, SourceAwareContext


class FunctionAccess(AST[Any]):
    def __init__(self, branch: List[AST]):
        self.branch = branch

    def execute(self, binding: ASTBinding) -> Any:
        current = None

        for item in self.branch:
            _binding = binding if current is None else SourceAwareContext(current, binding)
            current = item.execute(_binding)

        return current

    @staticmethod
    def parse(branch: List[AST]) -> 'FunctionAccess':
        return FunctionAccess(branch)



from typing import List

from scriptable.api import AST
from scriptable.api.ast_binding import ASTBinding
from scriptable.api.exit_value import ExitValue


class Return(AST[ExitValue]):
    def __init__(self, item: AST):
        self.item = item

    def execute(self, binding: ASTBinding) -> ExitValue:
        value = self.item.execute(binding)
        return ExitValue(0, value)

    @staticmethod
    def parse(branch: List[AST]) -> 'Return':
        return Return(branch[0])
